use core::cell::Cell;
use crate::cell::CellView;

pub mod input;
pub mod output;
mod conversion;

// Initialisation will always be async and won't complete until a state is available for all
// stateful transducers.
pub trait Stateful {
    type Value: Copy;

    fn state_cell(&self) -> CellView<Self::Value>;

    fn state(&self) -> Self::Value;
}

pub struct State<T: Copy> {
    state_cell: Cell<T>,
}

impl<T: Copy> State<T> {
    #[inline(always)]
    pub fn new(state_cell: Cell<T>) -> Self {
        Self { state_cell }
    }

    #[inline(always)]
    pub fn update(&self, value: T) {
        self.state_cell.set(value);
    }
}

impl<T: Copy> Stateful for State<T> {
    type Value = T;

    #[inline(always)]
    fn state_cell(&self) -> CellView<Self::Value> {
        (&self.state_cell).into()
    }

    #[inline(always)]
    fn state(&self) -> Self::Value {
        self.state_cell.get()
    }
}

impl<T: Copy> From<Cell<T>> for State<T> {
    #[inline(always)]
    fn from(state_cell: Cell<T>) -> Self {
        State::new(state_cell)
    }
}
